/*
 *  extend.c - subroutines for aplayex
 *
 *  Copyright (c) 2011 by NAKATA Hiroshi <webmaster@nakata-jp.org>
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <asm/byteorder.h>

#include "extend.h"

_formatinfo format_info[] = {
  { "signed 8bit", 8, 8},
  { "unsigned 8bit", 8, 8},
  { "signed 16bit Little Endian", 16, 16},
  { "signed 16bit Big Endian", 16, 16},
  { "unsigned 16bit Little Endian", 16, 16},
  { "unsigned 16bit Big Endian", 16, 16},
  { "signed 24bit Little Endian using low 3bytes in 32bit word", 24, 32},
  { "signed 24bit Big Endian using low 3bytes in 32bit word", 24, 32},
  { "unsigned 24bit Little Endian using low 3bytes in 32bit word", 24, 32},
  { "unsigned 24bit Big Endian using low 3bytes in 32bit word", 24, 32},
  { "signed 32bit Little Endian", 32, 32},
  { "signed 32bit Big Endian", 32, 32},
  { "unsigned 32bit Little Endian", 32, 32},
  { "unsigned 32bit Big Endian", 32, 32},
  { "float 32bit Little Endian, Range -1.0 to 1.0", 24, 32},
  { "float 32bit Big Endian, Range -1.0 to 1.0", 24, 32},
  { "float 64bit Little Endian, Range -1.0 to 1.0", 56, 64},
  { "float 64bit Big Endian, Range -1.0 to 1.0", 56, 64},
  { "IEC-958 Little Endian", 0, 0},
  { "IEC-958 Big Endian", 0, 0},
  { "Mu-Law", 0, 0},
  { "A-Law", 0, 0},
  { "Ima-ADPCM", 0, 0},
  { "MPEG", 0, 0},
  { "GSM", 0, 0},
  { "unknown", 0, 0},
  { "unknown", 0, 0},
  { "unknown", 0, 0},
  { "unknown", 0, 0},
  { "unknown", 0, 0},
  { "unknown", 0, 0},
  { "Special", 0, 0},
  { "signed 24bit Little Endian in 3bytes format", 24, 24},
  { "signed 24bit Big Endian in 3bytes format", 24, 24 },
  { "unsigned 24bit Little Endian in 3bytes format", 24, 24},
  { "unsigned 24bit Big Endian in 3bytes format", 24, 24},
  { "signed 20bit Little Endian in 3bytes format", 20, 24},
  { "signed 20bit Big Endian in 3bytes format", 20, 24},
  { "unsigned 20bit Little Endian in 3bytes format", 20, 24},
  { "unsigned 20bit Big Endian in 3bytes format", 20, 24},
  { "signed 18bit Little Endian in 3bytes format", 18, 24},
  { "signed 18bit Big Endian in 3bytes format", 18, 24},
  { "unsigned 18bit Little Endian in 3bytes format", 18, 24},
  { "unsigned 18bit Big Endian in 3bytes format", 18, 24}
};

int snd_pcm_hw_params_set_format_ex(snd_pcm_t *handle,
				    snd_pcm_hw_params_t *params,
				    struct _hwparams *hwparams,
				    const int quiet_mode,
				    const snd_pcm_stream_t stream)
{
	int i;
	int found_bit_depth = 0;

	hwparams->device_format = -1; /* set undefined mark */
	/* set the sample format */
	if ( snd_pcm_hw_params_test_format(handle, params, hwparams->format) == 0 ) {
		hwparams->device_format = hwparams->format;
		return snd_pcm_hw_params_set_format(handle, params,
					    	hwparams->device_format);
	}
		
	for ( i = 0; i <= SND_PCM_FORMAT_LAST; i++ ) {
		if ( format_info[i].bit_depth < found_bit_depth ) {
			continue;
		}
		if ( snd_pcm_hw_params_test_format(handle, params, i) != 0 ) {
			continue; /* device does not support that foramat */
		}
		found_bit_depth = format_info[i].bit_depth;
		hwparams->device_format = i; /* format decided */
	}
	if ( hwparams->device_format < 0 ) { 
		return -1;	/* format undecided */
	}
	if ( !quiet_mode &&
	     hwparams->device_format != hwparams->format ) {
		fprintf(stderr, "I/O format is \"%s\".\n",
		       format_info[hwparams->format].name);
		fprintf(stderr, "Device format is \"%s\".\n",
		       format_info[hwparams->device_format].name);
	}
	return snd_pcm_hw_params_set_format(handle, params,
					    hwparams->device_format);
}

unsigned int extended_len(const struct _hwparams *hwparams,
			  const unsigned int source_len)
{
  return (source_len * format_info[hwparams->device_format].bit_pack
	  / format_info[hwparams->format].bit_pack);
}

unsigned int unextended_len(const struct _hwparams *hwparams,
			    const unsigned int source_len)
{
  return (source_len * format_info[hwparams->format].bit_pack
	  / format_info[hwparams->device_format].bit_pack);
}

int convert_buffer(const int input_format,
		   const int output_format,
		   const int channels,
		   u_char *device_buf,
		   const u_char *read_buf,
		   const int count)
{
	int i;
	long l;
	float f;
	double d;
	u_char *p_dst = device_buf;
	const u_char *p_src = read_buf;
	u_char *p_data = (u_char *)&l;

	/* no need to convert, only copy */
	if ( input_format == output_format ) {
		i = count * channels * format_info[input_format].bit_pack / 8;
		memcpy(device_buf, read_buf, i);
		return (i);
	}

	/* loop for each sample */
	for ( i = 0; i < count * channels; i++ ) {
		/* convert input data to l & f */
		switch ( input_format ) {
		case SND_PCM_FORMAT_U8:
			l = *p_src;
			l -= 0x80;
			l <<= 24;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_S16_LE:
			l = __le16_to_cpu(*(__le16 *)p_src);
			l <<= 16;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_U16_LE:
			l = __le16_to_cpu(*(__le16 *)p_src);
			l <<= 16;
			l -= 0x80000000;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_S16_BE:
			l = __be16_to_cpu(*(__be16 *)p_src);
			l <<= 16;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_U16_BE:
			l = __be16_to_cpu(*(__be16 *)p_src);
			l <<= 16;
			l -= 0x80000000;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_S32_LE:
			l = __le32_to_cpu(*(__le32 *)p_src);
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_U32_LE:
			l = __le32_to_cpu(*(__le32 *)p_src);
			l -= 0x80000000;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_S32_BE:
			l = __be32_to_cpu(*(__be32 *)p_src);
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_U32_BE:
			l = __be32_to_cpu(*(__be32 *)p_src);
			l -= 0x80000000;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_S24_LE:
			l = __le32_to_cpu(*(__le32 *)p_src);
			l <<= 8;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_U24_LE:
			l = __le32_to_cpu(*(__le32 *)p_src);
			l <<= 8;
			l -= 0x80000000;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_S24_BE:
			l = __be32_to_cpu(*(__be32 *)p_src);
			l <<= 8;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_U24_BE:
			l = __be32_to_cpu(*(__be32 *)p_src);
			l <<= 8;
			l -= 0x80000000;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_S24_3LE:
#ifdef	__LITTLE_ENDIAN
			p_data[0] = 0;
			p_data[1] = p_src[0];
			p_data[2] = p_src[1];
			p_data[3] = p_src[2];
#else
			p_data[3] = 0;
			p_data[2] = p_src[0];
			p_data[1] = p_src[1];
			p_data[0] = p_src[2];
#endif
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_U24_3LE:
#ifdef	__LITTLE_ENDIAN
			p_data[0] = 0;
			p_data[1] = p_src[0];
			p_data[2] = p_src[1];
			p_data[3] = p_src[2];
#else
			p_data[3] = 0;
			p_data[2] = p_src[0];
			p_data[1] = p_src[1];
			p_data[0] = p_src[2];
#endif
			l -= 0x80000000;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_S24_3BE:
#ifdef	__LITTLE_ENDIAN
			p_data[0] = 0;
			p_data[1] = p_src[2];
			p_data[2] = p_src[1];
			p_data[3] = p_src[0];
#else
			p_data[3] = 0;
			p_data[2] = p_src[2];
			p_data[1] = p_src[1];
			p_data[0] = p_src[0];
#endif
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_U24_3BE:
#ifdef	__LITTLE_ENDIAN
			p_data[0] = 0;
			p_data[1] = p_src[2];
			p_data[2] = p_src[1];
			p_data[3] = p_src[0];
#else
			p_data[3] = 0;
			p_data[2] = p_src[2];
			p_data[1] = p_src[1];
			p_data[0] = p_src[0];
#endif
			l -= 0x80000000;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_S18_3LE:
#ifdef	__LITTLE_ENDIAN
			p_data[0] = 0;
			p_data[1] = p_src[0];
			p_data[2] = p_src[1];
			p_data[3] = p_src[2];
#else
			p_data[3] = 0;
			p_data[2] = p_src[0];
			p_data[1] = p_src[1];
			p_data[0] = p_src[2];
#endif
			l <<= 6;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_U18_3LE:
#ifdef	__LITTLE_ENDIAN
			p_data[0] = 0;
			p_data[1] = p_src[0];
			p_data[2] = p_src[1];
			p_data[3] = p_src[2];
#else
			p_data[3] = 0;
			p_data[2] = p_src[0];
			p_data[1] = p_src[1];
			p_data[0] = p_src[2];
#endif
			l <<= 6;
			l -= 0x80000000;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_S18_3BE:
#ifdef	__LITTLE_ENDIAN
			p_data[0] = 0;
			p_data[1] = p_src[2];
			p_data[2] = p_src[1];
			p_data[3] = p_src[0];
#else
			p_data[3] = 0;
			p_data[2] = p_src[2];
			p_data[1] = p_src[1];
			p_data[0] = p_src[0];
#endif
			l <<= 6;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_U18_3BE:
#ifdef	__LITTLE_ENDIAN
			p_data[0] = 0;
			p_data[1] = p_src[2];
			p_data[2] = p_src[1];
			p_data[3] = p_src[0];
#else
			p_data[3] = 0;
			p_data[2] = p_src[2];
			p_data[1] = p_src[1];
			p_data[0] = p_src[0];
#endif
			l <<= 6;
			l -= 0x80000000;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_S20_3LE:
#ifdef	__LITTLE_ENDIAN
			p_data[0] = 0;
			p_data[1] = p_src[0];
			p_data[2] = p_src[1];
			p_data[3] = p_src[2];
#else
			p_data[3] = 0;
			p_data[2] = p_src[0];
			p_data[1] = p_src[1];
			p_data[0] = p_src[2];
#endif
			l <<= 4;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_U20_3LE:
#ifdef	__LITTLE_ENDIAN
			p_data[0] = 0;
			p_data[1] = p_src[0];
			p_data[2] = p_src[1];
			p_data[3] = p_src[2];
#else
			p_data[3] = 0;
			p_data[2] = p_src[0];
			p_data[1] = p_src[1];
			p_data[0] = p_src[2];
#endif
			l <<= 4;
			l -= 0x80000000;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_S20_3BE:
#ifdef	__LITTLE_ENDIAN
			p_data[0] = 0;
			p_data[1] = p_src[2];
			p_data[2] = p_src[1];
			p_data[3] = p_src[0];
#else
			p_data[3] = 0;
			p_data[2] = p_src[2];
			p_data[1] = p_src[1];
			p_data[0] = p_src[0];
#endif
			l <<= 4;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_U20_3BE:
#ifdef	__LITTLE_ENDIAN
			p_data[0] = 0;
			p_data[1] = p_src[2];
			p_data[2] = p_src[1];
			p_data[3] = p_src[0];
#else
			p_data[3] = 0;
			p_data[2] = p_src[2];
			p_data[1] = p_src[1];
			p_data[0] = p_src[0];
#endif
			l <<= 4;
			l -= 0x80000000;
			d = l / (double)0x80000000;
			break;
		case SND_PCM_FORMAT_FLOAT_LE:
			l = __le32_to_cpu(*(__le32 *)p_src);
			f = *(float *)(&l);
			d = (double)f;
			l = d * 0x80000000;
			break;
		case SND_PCM_FORMAT_FLOAT_BE:
			l = __be32_to_cpu(*(__be32 *)p_src);
			f = *(float *)(&l);
			d = f;
			l = d * 0x80000000;
			break;
		case SND_PCM_FORMAT_MU_LAW:
		  	fprintf(stderr,
			       "This program does'nt support Mu-Law format.\n");
			prg_exit(EXIT_FAILURE);
		default:
			fprintf(stderr, "Unsupported input format.\n");
			prg_exit(EXIT_FAILURE);
		}
		p_src += (format_info[input_format].bit_pack / 8);
		f = (float)d;

		/* convert from l & f to output format */
		switch ( output_format ) {
		case SND_PCM_FORMAT_U8:
			*p_dst = ((l >> 24) - 0x80);
			break;
		case SND_PCM_FORMAT_U16_LE:
			l += 0x80000000;	/* fall through */
		case SND_PCM_FORMAT_S16_LE:
			*(int16_t *)p_dst = __cpu_to_le16(l>>16);
			break;
		case SND_PCM_FORMAT_U24_LE:
			l += 0x80000000;	/* fall through */
		case SND_PCM_FORMAT_S24_LE:
			*(int32_t *)p_dst = __cpu_to_le32(l);
			break;
		case SND_PCM_FORMAT_U32_LE:
			l += 0x80000000;	/* fall through */
		case SND_PCM_FORMAT_S32_LE:
			*((int32_t *)p_dst) = __cpu_to_le32(l);
			break;
		case SND_PCM_FORMAT_FLOAT_LE:
			*((int32_t *)p_dst) = __cpu_to_le32(*(int32_t *)&f);
			break;
		case SND_PCM_FORMAT_FLOAT64_LE:
			*(int64_t *)p_dst = __cpu_to_le64(*(int64_t *)&d);
			break;
		case SND_PCM_FORMAT_U24_3LE:
			l += 0x80000000;	/* fall through */
		case SND_PCM_FORMAT_S24_3LE:
			p_dst[0] = (l >> 8) & 0xff;
			p_dst[1] = (l >> 16) & 0xff;
			p_dst[2] = (l >> 24) & 0xff;
			break;
		case SND_PCM_FORMAT_U20_3LE:
			l += 0x80000000;	/* fall through */
		case SND_PCM_FORMAT_S20_3LE:
			l >>= 4;
			if ( l & 0x8000000 ) {
				l |= 0xf0000000;
			}
			p_dst[0] = (l & 0xff00) >> 8;
			p_dst[1] = (l & 0xff0000) >> 16;
			p_dst[2] = (l & 0xff000000) >> 24;
			break;
		case SND_PCM_FORMAT_U18_3LE:
			l += 0x80000000;	/* fall through */
		case SND_PCM_FORMAT_S18_3LE:
			l >>= 6;
			if ( l & 0x4000000 ) {
				l |= 0xfc000000;
			}
			p_dst[0] = (l & 0xff00) >> 8;
			p_dst[1] = (l & 0xff0000) >> 16;
			p_dst[2] = (l & 0xff000000) >> 24;
			break;
		case SND_PCM_FORMAT_U16_BE:
			l += 0x80000000;	/* fall through */
		case SND_PCM_FORMAT_S16_BE:
			*(int16_t *)p_dst = __cpu_to_be16(l>>16);
			break;
		case SND_PCM_FORMAT_U24_BE:
			l += 0x80000000;	/* fall through */
		case SND_PCM_FORMAT_S24_BE:
			*(int32_t *)p_dst = __cpu_to_be32(l);
			break;
		case SND_PCM_FORMAT_U32_BE:
			l += 0x80000000;	/* fall through */
		case SND_PCM_FORMAT_S32_BE:
			*(int32_t *)p_dst = __cpu_to_be32(l);
			break;
		case SND_PCM_FORMAT_U24_3BE:
			l += 0x80000000;	/* fall through */
		case SND_PCM_FORMAT_S24_3BE:
			p_dst[0] = (l & 0xff000000) >> 24;
			p_dst[1] = (l & 0xff0000) >> 16;
			p_dst[2] = (l & 0xff00) >> 8;
			break;
		case SND_PCM_FORMAT_U20_3BE:
			l += 0x80000000;	/* fall through */
		case SND_PCM_FORMAT_S20_3BE:
			l >>= 4;
			if ( l & 0x8000000 ) {
				l |= 0xf0000000;
			}
			p_dst[0] = (l & 0xff000000) >> 24;
			p_dst[1] = (l & 0xff0000) >> 16;
			p_dst[2] = (l & 0xff00) >> 8;
			break;
		case SND_PCM_FORMAT_U18_3BE:
			l += 0x80000000;	/* fall through */
		case SND_PCM_FORMAT_S18_3BE:
			l >>= 6;
			if ( l & 0x4000000 ) {
				l |= 0xfc000000;
			}
			p_dst[0] = (l & 0xff000000) >> 24;
			p_dst[1] = (l & 0xff0000) >> 16;
			p_dst[2] = (l & 0xff00) >> 8;
			break;
		case SND_PCM_FORMAT_FLOAT_BE:
			*(int32_t *)p_dst = __cpu_to_be32(*(int32_t *)&f);
			break;
		case SND_PCM_FORMAT_FLOAT64_BE:
			*(int64_t *)p_dst = __cpu_to_be64(*(int64_t *)&d);
			break;
		  
		case SND_PCM_FORMAT_IEC958_SUBFRAME_LE:
		case SND_PCM_FORMAT_IEC958_SUBFRAME_BE:
		case SND_PCM_FORMAT_MU_LAW:
		case SND_PCM_FORMAT_A_LAW:
		case SND_PCM_FORMAT_IMA_ADPCM:
		case SND_PCM_FORMAT_MPEG:
		case SND_PCM_FORMAT_GSM:
		case SND_PCM_FORMAT_SPECIAL:
		default:
			fprintf(stderr,
				"Unsupported output format \"%s\".\nSorry.\n",
				format_info[output_format].name);
			prg_exit(EXIT_FAILURE);
		}
		p_dst += (format_info[output_format].bit_pack / 8);
	}
	return (i * format_info[output_format].bit_pack / 8);
}
